@@ -3,6 +3,11 @@
 """Functional tests for ``pyramid_simpleauth``."""
 
 import unittest
+from mock import Mock
+
+from pyramid_basemodel import Session
+from pyramid_simpleauth import model
+import transaction
 
 try: # pragma: no cover
     from webtest import TestApp
@@ -33,19 +38,59 @@ def config_factory(**settings):
     return config
 
 
-class TestSignup(unittest.TestCase):
+class BaseTestCase(unittest.TestCase):
+
     def setUp(self):
         """Configure the Pyramid application."""
         
-        self.config = config_factory()
+        # Configure redirect routes
+        view_names = ('change_password', 'change_username', 'confirm_email',
+                      'prefer_email')
+        settings = dict(('simpleauth.after_%s_route' % view_name,
+                         'success_path') for view_name in view_names)
+        self.config = config_factory(**settings)
+        # Add routes for change_password, change_username, confirm_email_address
+        # and preferred_email
+        self.config.add_route('success_path', 'victory_path')
         self.app = TestApp(self.config.make_wsgi_app())
     
     def tearDown(self):
         """Make sure the session is cleared between tests."""
         
-        from pyramid_basemodel import Session
         Session.remove()
     
+    def makeUser(self, username, password):
+        """Create and save a user with the credentials provided."""
+        
+        user = model.User()
+        user.username = username
+        user.password = model.encrypt(password)
+        model.save(user)
+        transaction.commit()
+        Session.add(user)
+        return user
+
+    def makeUserWithEmail(self):
+        "Helper method that creates a user with an email"
+        user = self.makeUser(u'thruflo', u'password')
+        Session.add(user)
+        user.emails.append(model.Email(address=u'foo@example.com'))
+        transaction.commit()
+        Session.add(user)
+        return user
+
+    def authenticate(self, user):
+        "Authenticate user"
+        post_data = {
+            'username': 'thruflo',
+            'password': 'password'
+        }
+        headers = {'X-Requested-With': 'XMLHttpRequest'}
+        return self.app.post('/auth/authenticate', post_data, headers=headers)
+
+
+class TestSignup(BaseTestCase):
+    
     def test_render_signup_form(self):
         """A GET request to the signup view should render the signup form."""
         
@@ -134,7 +179,6 @@ class TestSignup(unittest.TestCase):
     def test_signup_event(self):
         """Signup fires a ``UserSignedUp`` event."""
         
-        from mock import Mock
         from pyramid_simpleauth.events import UserSignedUp
         from pyramid_simpleauth.model import User
         # Setup event listener.
@@ -156,29 +200,7 @@ class TestSignup(unittest.TestCase):
         self.assertTrue(isinstance(event.user, User))
     
 
-class TestLogin(unittest.TestCase):
-    def setUp(self):
-        """Configure the Pyramid application."""
-        
-        self.config = config_factory()
-        self.app = TestApp(self.config.make_wsgi_app())
-    
-    def tearDown(self):
-        """Make sure the session is cleared between tests."""
-        
-        from pyramid_basemodel import Session
-        Session.remove()
-    
-    def makeUser(self, username, password):
-        """Create and save a user with the credentials provided."""
-        
-        import transaction
-        from pyramid_simpleauth import model
-        user = model.User()
-        user.username = username
-        user.password = model.encrypt(password)
-        model.save(user)
-        transaction.commit()
+class TestLogin(BaseTestCase):
     
     def test_render_login_form(self):
         """A GET request to the login view should render the login form."""
@@ -238,7 +260,7 @@ class TestLogin(unittest.TestCase):
         self.config.add_route('index', 'some/path')
         self.app = TestApp(self.config.make_wsgi_app())
         res = self.app.post('/auth/login', post_data, status=302)
-        self.assertTrue(res.headers['Location'] == 'http://localhost/some/path')
+        self.assertEquals(res.location, 'http://localhost/some/path')
         # Or the `simpleauth.after_logout_route` route if specified and exposed.
         settings = {
             'simpleauth.after_login_route': 'flobble'
@@ -252,7 +274,6 @@ class TestLogin(unittest.TestCase):
     def test_login_event(self):
         """Login fires a ``UserLoggedIn`` event."""
         
-        from mock import Mock
         from pyramid_simpleauth.events import UserLoggedIn
         from pyramid_simpleauth.model import User
         # Setup event listener.
@@ -273,29 +294,7 @@ class TestLogin(unittest.TestCase):
         self.assertTrue(isinstance(event.user, User))
     
 
-class TestAuthenticate(unittest.TestCase):
-    def setUp(self):
-        """Configure the Pyramid application."""
-        
-        self.config = config_factory()
-        self.app = TestApp(self.config.make_wsgi_app())
-    
-    def tearDown(self):
-        """Make sure the session is cleared between tests."""
-        
-        from pyramid_basemodel import Session
-        Session.remove()
-    
-    def makeUser(self, username, password):
-        """Create and save a user with the credentials provided."""
-        
-        import transaction
-        from pyramid_simpleauth import model
-        user = model.User()
-        user.username = username
-        user.password = model.encrypt(password)
-        model.save(user)
-        transaction.commit()
+class TestAuthenticate(BaseTestCase):
     
     def test_authenticate_requires_xhr(self):
         """Authenticate must be called with an XMLHTTPRequest."""
@@ -340,7 +339,6 @@ class TestAuthenticate(unittest.TestCase):
     def test_authenticate_logged_in_event(self):
         """Authenticate fires a ``UserLoggedIn`` event."""
         
-        from mock import Mock
         from pyramid_simpleauth.events import UserLoggedIn
         from pyramid_simpleauth.model import User
         # Setup event listener.
@@ -362,29 +360,7 @@ class TestAuthenticate(unittest.TestCase):
         self.assertTrue(isinstance(event.user, User))
     
 
-class TestLogout(unittest.TestCase):
-    def setUp(self):
-        """Configure the Pyramid application."""
-        
-        self.config = config_factory()
-        self.app = TestApp(self.config.make_wsgi_app())
-    
-    def tearDown(self):
-        """Make sure the session is cleared between tests."""
-        
-        from pyramid_basemodel import Session
-        Session.remove()
-    
-    def makeUser(self, username, password):
-        """Create and save a user with the credentials provided."""
-        
-        import transaction
-        from pyramid_simpleauth import model
-        user = model.User()
-        user.username = username
-        user.password = model.encrypt(password)
-        model.save(user)
-        transaction.commit()
+class TestLogout(BaseTestCase):
     
     def test_logout(self):
         """Logout forgets the user."""
@@ -428,7 +404,6 @@ class TestLogout(unittest.TestCase):
     def test_loggedout_event(self):
         """Logout fires a ``UserLoggedOut`` event."""
         
-        from mock import Mock
         from pyramid_simpleauth.events import UserLoggedOut
         from pyramid_simpleauth.model import User
         # Setup event listener.
@@ -455,7 +430,6 @@ class TestLogout(unittest.TestCase):
           authenticated user.
         """
         
-        from mock import Mock
         from pyramid_simpleauth.events import UserLoggedOut
         # Setup event listener.
         mock_subscriber = Mock()
@@ -468,3 +442,290 @@ class TestLogout(unittest.TestCase):
         self.assertFalse(mock_subscriber.called)
     
 
+class TestChangePassword(BaseTestCase):
+
+    def test_wrong_old_password(self):
+        "No password change if old password is not corret"
+
+        # Create a user.
+        user = self.makeUser('thruflo', 'password')
+        Session.add(user)
+        old_hash = user.password
+
+        self.authenticate(user)
+
+        # Attempt to change password.
+        post_data = {
+            'old_password': 'foobarbaz',
+            'new_password': 'swordpas',
+            'new_confirm':  'swordpas',
+            'next':         '/foo/bar',
+        }
+        res = self.app.post('/auth/change_password', post_data)
+
+        # Verify that password hasn't changed
+        Session.add(user)
+        Session.refresh(user)
+        self.assertTrue("Wrong current password" in res.body)
+        self.assertTrue("/foo/bar" in res.body)
+        self.assertEquals(user.password, old_hash)
+
+    def test_wrong_old_password_returns_valid_user(self):
+        "Bug fix: user template param must not be None"
+        from pyramid_simpleauth.view import change_password
+        from pyramid import testing
+
+        # Create a user.
+        user = self.makeUser('thruflo', 'password')
+        Session.add(user)
+
+        post_data = {
+            'old_password': 'foobarbaz',
+            'new_password': 'swordpas',
+            'new_confirm':  'swordpas',
+        }
+        request = testing.DummyRequest(post=post_data)
+        request.user = user
+        testing.setUp(settings={})
+        res = change_password(request)
+
+        self.assertTrue(res['user'])
+
+
+    def test_new_passwords_dont_match(self):
+        "No password change if new passwords don't match"
+
+        # Create a user.
+        user = self.makeUser('thruflo', 'password')
+        Session.add(user)
+        old_hash = user.password
+
+        self.authenticate(user)
+
+        # Attempt to change password.
+        post_data = {
+            'old_password': 'password',
+            'new_password': 'swordpas',
+            'new_confirm':  'oswdrpsa',
+        }
+        res = self.app.post('/auth/change_password', post_data)
+
+        # Verify that password hasn't changed
+        Session.add(user)
+        Session.refresh(user)
+        self.assertTrue("Fields do not match" in res.body)
+        self.assertEquals(user.password, old_hash)
+
+    def test_sucess(self):
+        "If all conditions are met, change password"
+
+        # Create a user.
+        user = self.makeUser('thruflo', 'password')
+        Session.add(user)
+        old_hash = user.password
+
+        self.authenticate(user)
+
+        # Attempt to change password.
+        post_data = {
+            'old_password': 'password',
+            'new_password': 'swordpas',
+            'new_confirm':  'swordpas',
+            'next':         '/foo/bar',
+        }
+        res = self.app.post('/auth/change_password', post_data)
+
+        # Verify that password has changed
+        Session.add(user)
+        Session.refresh(user)
+        self.assertNotEquals(user.password, old_hash)
+
+        # Verify redirect
+        self.assertEquals(res.headers['Location'], 'http://localhost/foo/bar')
+
+
+class TestChangeUsername(BaseTestCase):
+
+    def test_success(self):
+        "Change username with valid input"
+        # Create a user.
+        user = self.makeUser(u'thruflo', u'password')
+
+        # Attempt to change username
+        post_data = {
+            'username': u'bob',
+            'next':     '/foo/bar',
+        }
+        self.authenticate(user)
+        res = self.app.post('/auth/change_username', post_data)
+
+        # Verify redirect
+        self.assertEquals(res.location, 'http://localhost/foo/bar')
+
+        # Verify that username has changed
+        Session.add(user)
+        self.assertEquals(user.username, 'bob')
+
+    def test_failure(self):
+        "Change username with invalid input"
+        # Create a user.
+        user = self.makeUser(u'thruflo', u'password')
+
+        # Attempt to assign bogus username
+        post_data = {
+            'username': u'$ @ 88 , /',
+            'next':     '/foo/bar',
+        }
+        self.authenticate(user)
+        res = self.app.post('/auth/change_username', post_data)
+
+        # Verify response body
+        self.assertTrue('No spaces or funny characters' in res.body)
+        self.assertTrue('/foo/bar' in res.body,
+                        "Response body should contain next hidden field")
+
+        # Verify that username has not changed
+        Session.add(user)
+        self.assertEquals(user.username, 'thruflo')
+
+
+class TestConfirmEmailAddress(BaseTestCase):
+
+    def makeConfirmationLink(self, email):
+        "Helper method that makes a valid confirmation link"
+        from pyramid_simpleauth.model import get_confirmation_link
+        request = Mock()
+        request.route_url.return_value = '/auth/confirm'
+        return get_confirmation_link(request, email)
+
+    def test_success(self):
+        "Token is valid, email address should be confirmed"
+        # Create a user
+        user = self.makeUserWithEmail()
+
+        # Sanity check
+        self.assertFalse(user.emails[0].is_confirmed)
+
+        # Get valid confirmation link
+        email = user.emails[0]
+        confirmation_link = self.makeConfirmationLink(email)
+
+        # Attempt to confirm email address
+        res = self.app.get(confirmation_link)
+        self.assertTrue(res.location.endswith('victory_path'))
+
+        # Now configure settings with a route that doesn't exist
+        settings = {'simpleauth.after_confirm_email_route': 'success_path'}
+        self.config = config_factory(**settings)
+        # Not adding the route!
+        self.app = TestApp(self.config.make_wsgi_app())
+        res = self.app.get(confirmation_link)
+        self.assertEquals(res.location, 'http://localhost/')
+
+        # Verify that email address has been confirmed
+        Session.add(email)
+        Session.refresh(email)
+        self.assertTrue(email.is_confirmed)
+
+    def test_failure(self):
+        "Token is invalid, email address should not be confirmed"
+        # Create a user
+        user = self.makeUserWithEmail()
+
+        # Sanity check
+        self.assertFalse(user.emails[0].is_confirmed)
+
+        # Bogus attempts to confirm email address
+        # 1. malformed link
+        url = '/auth/confirm/foo'
+        res = self.app.get(url)
+        self.assertTrue('invalid' in res.body)
+
+        # 2. invalid token
+        email = user.emails[0]
+        url = self.makeConfirmationLink(email) + 'gibberish'
+        res = self.app.get(url)
+        self.assertTrue('invalid' in res.body)
+
+
+        # Verify that email address has been confirmed
+        Session.add(email)
+        Session.refresh(email)
+        self.assertFalse(email.is_confirmed)
+
+
+class TestPreferEmail(BaseTestCase):
+
+    def test_success(self):
+        "Set preferred email address"
+        # Create user with email address
+        user = self.makeUserWithEmail()
+        # Add another one
+        user.emails.append(model.Email(address=u'bar@example.com',
+                           is_preferred=True))
+        model.save(user)
+        transaction.commit()
+        Session.add(user)
+
+        email1, email2 = user.emails
+
+        # Sanity check
+        self.assertNotEquals(user.preferred_email, email1)
+        self.assertEquals(user.preferred_email, email2)
+
+        # Attempt to make the address primary
+        self.authenticate(user)
+        self.app.post('/auth/prefer_email', {
+            'email_address': email1.address
+        })
+
+        # Verify that email is not the user's preferred email
+        Session.add(email1)
+        Session.refresh(email1)
+        self.assertEquals(user.preferred_email, email1)
+        self.assertNotEquals(user.preferred_email, email2)
+
+    def test_failure(self):
+        "Attempt to set preferred email address with invalid input"
+        # Create user with email address
+        user = self.makeUserWithEmail()
+        email = user.emails[0]
+
+        # Attempt to make the address primary
+        self.authenticate(user)
+        self.app.post('/auth/prefer_email', {
+            'email_address': email.address + 'not an email address'
+        })
+
+        # Email address should not be prefered
+        self.assertNotEquals(user.preferred_email, email)
+
+    def test_prefer_non_persisted_email(self):
+        "Set non-persisted email object as new preferred email"
+        # Create user without any email address
+        user = self.makeUser('bob', '123')
+
+        # Directly set new email as preferred email
+        email = model.Email(address=u'bar@example.com')
+        user.preferred_email = email
+        model.save(user)
+        transaction.commit()
+
+        # Verify that the new email is now the preferred email
+        Session.add(user)
+        self.assertEquals(user.preferred_email.address, u'bar@example.com')
+
+    def test_preferred_if_only_one(self):
+        "If user has only one email, consider it as preferred email"
+        # Create user without any email address
+        user = self.makeUser('bob', '123')
+
+        # Directly set new email as preferred email
+        email = model.Email(address=u'bar@example.com')
+        user.emails.append(email)
+        model.save(user)
+        transaction.commit()
+
+        # Verify that the new email is now the preferred email
+        Session.add(user)
+        self.assertEquals(user.preferred_email.address, u'bar@example.com')
